/*******************************************************************************
 * Copyright (c) 2010 Freescale Semicondcutor and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Freescale Semicondcutor - Initial API and implementation
 *******************************************************************************/

// This is a simple multithreaded program which asks n threads to find as many
// prime numbers as they can in a second. The thread to calculate the most
// is declared the winner of the "race". Ten races are run.
//
// This program can be used to explore and test multi-context ideas. By setting
// a breakpoint in 'foundCount++', the debugger will likely end up stopping the
// program with all threads in user code (rather than in system code).
//
// This implementation is specifically for Windows, and requires MinGW.
//

#include <process.h>
#include <windows.h>
#include <iostream>

HANDLE event = 0;
volatile boolean g_done = false;
volatile boolean g_putPensDown = false;
DWORD g_winningThread = 0;	// threads: access only in critical section
int g_winningCount = 0;		// threads: access only in critical section
CRITICAL_SECTION g_critsec;
int g_threadCount = 5;

static void threadfunc(void* param)
{
	int tid = (int)GetCurrentThreadId();
	std::cout << "[TID " << tid << "] Thread created." << std::endl << std::flush;

	while (true)
	{
		WaitForSingleObject(event, INFINITE);
		if (g_done)
		{
			std::cout << "[TID " << tid << "] Thread closing shop." << std::endl << std::flush;
			return;
		}

		int i = 2;	// start with the first prime number
		int foundCount = 0;
		while (!g_putPensDown)
		{
			for (int divisor = i-1; divisor > 1 && !g_putPensDown; divisor--)
			{
				if (i % divisor == 0)
				{
					foundCount++;
					break;
				}
			}

			i++;
		}


		EnterCriticalSection(&g_critsec);
		if (foundCount > g_winningCount)
		{
			g_winningCount = foundCount;
			g_winningThread = tid;
		}
		LeaveCriticalSection(&g_critsec);
	}
}

int main(int argc, char *argv[])
{
	// Create synchronization objects
	event = CreateEvent(NULL, TRUE, FALSE, NULL);
	assert(event != NULL);
    InitializeCriticalSection(&g_critsec);

    // Create the threads
	for (int i = 0; i < g_threadCount; i++)
	{
		uintptr_t thandle = _beginthread(threadfunc, 0, NULL );
		assert(thandle != -1UL);
	}

	// Give the threads a chance to get ready
	Sleep(100);

    for (int i = 0; i < 10; i++)
    {
    	// reset race state
    	g_putPensDown = false;
    	g_winningThread = 0;
    	g_winningCount = 0;

    	std::cout << "----------------------------------" << std::endl << std::flush;
    	std::cout << "Waking up threads" << std::endl;
    	PulseEvent(event);

    	std::cout << "Giving threads one second to do their work." << std::endl << std::flush;
    	Sleep(1000);

    	std::cout << "Telling threads to put down their pens." << std::endl << std::flush;
    	g_putPensDown = true;

    	// Give the threads a chance to drop their pens
    	Sleep(100);

    	// Declare winner
		std::cout << "Thread " << g_winningThread << " won with " << g_winningCount << " prime numbers found." << std::endl << std::flush;

    	// Then start a new round...
    }

    // Wind things down
    g_done = true;
    PulseEvent(event);
	Sleep(100); // Give the threads a chance to close shop

	CloseHandle(event);
    DeleteCriticalSection(&g_critsec);
    return 0;
}
